#version 330
#define SM_SIZE 4096.0

//uncomment to get smoother shadows
//#define HQ_SHADOWS

uniform bool uni_HasDiffuseMap;
uniform bool uni_HasNormalMap;

uniform vec3 uni_DiffuseColor;
uniform vec3 uni_AmbientColor;
uniform vec3 uni_SpecularColor;

uniform sampler2D uni_DiffuseMap;
uniform sampler2DShadow uni_ShadowMap;
uniform sampler2D uni_NormalMap;
uniform vec3 uni_ParallelLightColor;
uniform float uni_Shininess;

in vec2 var_TexCoord;
//in vec3 var_Normal; //normals are in tangent space, default (0,0,1)
in vec3 var_EyeVec; //in TANGENT space, not normalized
in vec3 var_LightVec; //in TANGENT space, not normalized
in vec4 var_SMTexCoords;
in vec4 var_Pos;

out vec4 frag_Color;

const vec2 texmapscale = vec2(1.0/SM_SIZE,1.0/SM_SIZE);

float offset_lookup(const in sampler2DShadow map,
                    const in vec4 loc,
                    const in vec2 offset)
{
return textureProj(map, vec4(loc.xy + offset * texmapscale * loc.w,
                               loc.z, loc.w));
}

void main(void)
{
	vec4 baseColor = vec4(1,1,1,1);
	if(uni_HasDiffuseMap)
		baseColor = texture(uni_DiffuseMap,var_TexCoord);

	//simple alpha testing
	if(baseColor.a<0.1)
		discard;

	//do advanced PCF for shadows by using 16/4 samples (http://http.developer.nvidia.com/GPUGems/gpugems_ch11.html)
	

	#ifdef HQ_SHADOWS
	float sum = 0;
	float x, y;

	for (y = -1.5; y <= 1.5; y += 1.0)
	for (x = -1.5; x <= 1.5; x += 1.0)
		sum += offset_lookup(uni_ShadowMap,var_SMTexCoords,vec2(x,y));

	float shadowCoeff = sum/16.0;
	#else
	float sum = 0;
	float x, y;

	for (y = -0.5; y <= 0.5; y += 1.0)
	for (x = -0.5; x <= 0.5; x += 1.0)
		sum += offset_lookup(uni_ShadowMap,var_SMTexCoords,vec2(x,y));

	float shadowCoeff = sum/4.0;
	#endif	

/*
	//testing shadowmap dithering, not really ok yet
	vec2 offset = vec2(length(fract(var_Pos.xy * 0.5)) > 0.25);  // mod
	offset.y += offset.x;  // y ^= x in floating point

	if (offset.y > 1.1)
		offset.y = 0;

	float shadowCoeff = (
		offset_lookup(uni_ShadowMap, var_SMTexCoords, offset + vec2(-1.5, 0.5)) +
		offset_lookup(uni_ShadowMap, var_SMTexCoords, offset + vec2(0.5, 0.5)) +
		offset_lookup(uni_ShadowMap, var_SMTexCoords, offset + vec2(-1.5, -1.5)) +
		offset_lookup(uni_ShadowMap, var_SMTexCoords, offset + vec2(0.5, -1.5)) ) * 0.25;
*/

	//use tangent space default normal if no normal map - this should not happen anyway because only objects with normal maps are drawn with this shader by the game
	vec3 normal = vec3(0,0,1);
	if(uni_HasNormalMap)
	{
		normal = normalize(texture(uni_NormalMap, var_TexCoord).rgb * 2.0 - 1.0); //scale 0 to 1 intensity to -1 to 1 coords
		normal.y = -normal.y; //y has to be flipped because the texture was generated differently
	}

	//ambient light = ambientMaterial * ambientLight
	vec3 ambient = uni_AmbientColor * 0.2;

	vec3 vNormLightVec = normalize(var_LightVec);

	//diffuse light = DiffMaterial * DiffLight * Lambert
	float diff = max(dot(normal,vNormLightVec),0.0);
	//float diff = max(dot(normal,-var_LightDir),0.0);
	vec3 diffuse = diff * uni_ParallelLightColor * uni_DiffuseColor;

	//specular light = specMaterial * specLight * phong
	float spec = 0.0;
	if(diff>0.0) //only calc if light shines on this point
	{
		vec3 R = reflect(-vNormLightVec,normal);		
		vec3 E = normalize(var_EyeVec);

		spec = pow(max(dot(R,E),0.0),uni_Shininess);
	}
	vec3 specular = spec * uni_SpecularColor * 0.8;

	//final color = ambient + diffuse + specular
	frag_Color = vec4(ambient * baseColor.xyz + diffuse * baseColor.xyz * shadowCoeff + specular * shadowCoeff, baseColor.a);
}